""" This example illustrates the interface of the description logics component.
"""

import pymimir.advanced.formalism as formalism
import pymimir.advanced.search as search
import pymimir.advanced.datasets as datasets
import pymimir.advanced.languages.description_logics as description_logics

from pathlib import Path

ROOT_DIR = (Path(__file__).parent.parent.parent).absolute()


def main():
    domain_filepath = str(ROOT_DIR / "data" / "gripper" / "domain.pddl")
    problem_filepath = str(ROOT_DIR / "data" / "gripper" / "p-1-0.pddl")

    # Create some states
    search_context_options = search.SearchContextOptions()
    search_context_options.mode = search.SearchMode.GROUNDED
    generalized_search_context = search.GeneralizedSearchContext.create(domain_filepath, [problem_filepath], search_context_options)

    state_space_options = datasets.StateSpaceOptions()
    state_space_options.symmetry_pruning = True 
    generalized_state_space_options = datasets.GeneralizedStateSpaceOptions()
    knowledge_base_options = datasets.KnowledgeBaseOptions()
    knowledge_base_options.state_space_options = state_space_options
    knowledge_base_options.generalized_state_space_options = generalized_state_space_options
    knowledge_base = datasets.KnowledgeBase.create(generalized_search_context, knowledge_base_options)

    # Create a CNFGrammar for the domain.
    # You can also write down your own grammar in a BNF specification that will automatically be translated into a CNF representation.
    grammar = description_logics.CNFGrammar.create(description_logics.GrammarSpecificationEnum.FRANCES_ET_AL_AAAI2021, knowledge_base.get_domain())

    pruning_function = description_logics.StateListRefinementPruningFunction(knowledge_base.get_generalized_state_space())
    sentences = description_logics.GeneratedSentencesContainer()
    repositories = description_logics.Repositories()
    sentence_generator = description_logics.GeneratorVisitor(pruning_function, sentences, repositories, 5)
    sentence_generator.visit(grammar)

    print("================================================================================")
    print("Input Grammar in CNF")
    print("================================================================================")
    print(grammar)

    print("================================================================================")
    print("Generated concepts")
    print("================================================================================")
    for complexity, constructors in enumerate(sentences.get_concepts().get(grammar.get_concept_start_symbol())):
        print("Complexity:", complexity)
        for constructor in constructors:
            print(f"    {constructor}")

            assert(grammar.test_match(constructor))

    print("================================================================================")
    print("Generated roles")
    print("================================================================================")
    for complexity, constructors in enumerate(sentences.get_roles().get(grammar.get_role_start_symbol())):
        print("Complexity:", complexity)
        for constructor in constructors:
            print(f"    {constructor}")

            assert(grammar.test_match(constructor))

    print("================================================================================")
    print("Generated booleans")
    print("================================================================================")
    for complexity, constructors in enumerate(sentences.get_booleans().get(grammar.get_boolean_start_symbol())):
        print("Complexity:", complexity)
        for constructor in constructors:
            print(f"    {constructor}")

            assert(grammar.test_match(constructor))

    print("================================================================================")
    print("Generated numericals")
    print("================================================================================")
    for complexity, constructors in enumerate(sentences.get_numericals().get(grammar.get_numerical_start_symbol())):
        print("Complexity:", complexity)
        for constructor in constructors:
            print(f"    {constructor}")

            assert(grammar.test_match(constructor))

    print("================================================================================")
    print("Evaluate numericals")
    print("================================================================================")
    denotation_repositories = description_logics.DenotationRepositories()
    for complexity, constructors in enumerate(sentences.get_numericals().get(grammar.get_numerical_start_symbol())):
        print("Complexity:", complexity)
        for constructor in constructors:
            print(f"    {constructor}")
            for state_space in knowledge_base.get_state_spaces():
                for vertex in state_space.get_graph().get_vertices():
                    state = vertex.get_property_0()
                    problem = vertex.get_property_1()
                    context = description_logics.EvaluationContext(state, problem, denotation_repositories)
                    value = constructor.evaluate(context).get_data()
                    print("State:", state.to_string(problem))
                    print("Value:", value)
    
    

if __name__ == "__main__":
    main()



""" This example illustrates the interface of a generic static graph taking arbitrary python tuples as vertex and edge property.
"""

import pymimir.advanced.graphs as graphs


def main():
    graph = graphs.StaticPyGraph()
    v0 = graph.add_vertex((0,0))
    v1 = graph.add_vertex((0,1))
    v2 = graph.add_vertex((1,1))
    v3 = graph.add_vertex((1,0))
    graph.add_undirected_edge(v0, v1, ("a",1))
    graph.add_undirected_edge(v1, v2, ("b",2))
    graph.add_undirected_edge(v2, v3, ("c",3))
    graph.add_undirected_edge(v3, v0, ("d",4))

    print(graph)

    # Run basic graph algorithms with switchable edge direction

    edge_weights = [1] * graph.get_num_edges()
    source_indices = [0,1]

    graph.compute_strong_components()

    graph.compute_forward_breadth_first_search(source_indices)
    graph.compute_forward_depth_first_search(source_indices)
    graph.compute_forward_floyd_warshall_all_pairs_shortest_paths(edge_weights)
    graph.compute_forward_dijkstra_shortest_paths(edge_weights, source_indices)
    try:
        graph.compute_forward_topological_sort()  # graph is cyclic => this will throw ValueError
    except ValueError:
        pass
    
    graph.compute_backward_breadth_first_search(source_indices)
    graph.compute_backward_depth_first_search(source_indices)
    graph.compute_backward_floyd_warshall_all_pairs_shortest_paths(edge_weights)
    graph.compute_backward_dijkstra_shortest_paths(edge_weights, source_indices)
    try:
        graph.compute_backward_topological_sort()  # graph is cyclic => this will throw ValueError
    except ValueError:
        pass

    # Iterate over vertices
    for vertex in graph.get_vertices():
        # Access vertex properties through generic member
        vertex.get_index()      
        vertex.get_property_0()  # python tuple

        # Iterate over adjacent structures
        for adj_vertex in graph.get_forward_adjacent_vertices(vertex.get_index()):
            pass

        for adj_vertex in graph.get_backward_adjacent_vertices(vertex.get_index()):
            pass 

        for adj_edge in graph.get_forward_adjacent_edges(vertex.get_index()):
            pass 

        for adj_edge in graph.get_backward_adjacent_edges(vertex.get_index()):
            pass 

        for adj_v_idx in graph.get_forward_adjacent_vertex_indices(vertex.get_index()):
            pass 

        for adj_v_idx in graph.get_backward_adjacent_vertex_indices(vertex.get_index()):
            pass 

        for adj_e_idx in graph.get_forward_adjacent_edge_indices(vertex.get_index()):
            pass 

        for adj_e_idx in graph.get_backward_adjacent_edge_indices(vertex.get_index()):
            pass
            
    # Iterate over edges
    for edge in graph.get_edges():
        # Access edge properties through generic member
        edge.get_index()  
        edge.get_source()   
        edge.get_target()   
        edge.get_property_0()  # python tuple



if __name__ == "__main__":
    main()